"""
Shared constants across the various oper8.x tools
"""

# The key in the CR spec that holds the datastore information
SPEC_DATASTORES = "datastores"

# The key inside an individual spec.<datastore_type> that holds connection info
SPEC_DATASTORE_CONNECTION = "connection"

GLOBAL_SLOT = "global"

DEPS_ANNOTATION = "oper8.org/dependency-hash"

SERVICE_ACCOUNT_NAME_WORKLOAD_SCHEDULER = "workload-scheduler"

PULL_POLICY = "IfNotPresent"
